# -*- coding: utf-8 -*-
#
#  install.py - an installer module for ninix
#  Copyright (C) 2001, 2002 by Tamito KAJIYAMA
#  Copyright (C) 2002, 2003 by MATSUMURA Namihiko <nie@counterghost.net>
#  Copyright (C) 2002-2011 by Shyouzou Sugitani <shy@users.sourceforge.jp>
#  Copyright (C) 2003 by Shun-ichi TAHARA <jado@flowernet.gr.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#

import distutils.dir_util
import os
import shutil
import stat
import sys
import logging
import tempfile
import urllib
import zipfile

try:
    import gtk
except:
    gtk = None

import ninix.home
import ninix.version


# mode masks ## FIXME
INTERACTIVE = 1
DOWNLOAD    = 2

class URLopener(urllib.FancyURLopener): ## FIXME
    version = 'ninix-aya/%s' % ninix.version.VERSION

urllib._urlopener = URLopener() # XXX


class InstallError(Exception):
    pass


def fatal(error):
    raise InstallError, error


class Installer:

    def __init__(self, mode=INTERACTIVE|DOWNLOAD, gui=1):
        self.mode = mode
        if gui and gtk is not None:
            self.dialog = gtk.MessageDialog(type=gtk.MESSAGE_QUESTION,
                                            buttons=gtk.BUTTONS_YES_NO)
        else:
            self.dialog = None

    def check_archive(self, filename):
        # check archive format
        basename, ext = os.path.splitext(filename)
        ext = ext.lower()
        if ext in ['.nar', '.zip']:
            pass
        else:
            fatal('unknown archive format')

    def extract_files(self, filename):
        # extract files from the archive
        tmpdir = tempfile.mkdtemp('ninix-aya')
        shutil.rmtree(tmpdir) # XXX
        try:
            os.makedirs(tmpdir)
        except:
            fatal('cannot make temporary directory')
        url = None
        if filename.startswith('http:') or filename.startswith('ftp:'):
            url = filename
            filename = download(filename, tmpdir)
            if filename is None:
                shutil.rmtree(tmpdir)
                fatal('cannot download the archive file')
        zf = zipfile.ZipFile(filename)
        try:
            zf.extractall(tmpdir)
            zf.close()
        except:
            shutil.rmtree(tmpdir)
            fatal('cannot extract files from the archive')
        if url and not self.mode & DOWNLOAD: ## FIXME
            os.remove(filename)
        for (dirpath, dirnames, filenames) in os.walk(tmpdir):
            for name in dirnames:
                path = os.path.join(dirpath, name)
                st_mode = os.stat(path).st_mode
                os.chmod(path, st_mode | stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR)
            for name in filenames:
                path = os.path.join(dirpath, name)
                st_mode = os.stat(path).st_mode
                os.chmod(path, st_mode | stat.S_IWUSR | stat.S_IRUSR)
        self.rename_files(tmpdir)
        return tmpdir

    def get_file_type(self, tmpdir):
        # check the file type
        inst = ninix.home.read_install_txt(tmpdir)
        if not inst:
            fatal('cannot read install.txt from the archive')
        filetype = inst.get('type')
        if filetype == 'ghost':
            if os.path.exists(os.path.join(tmpdir, 'ghost', 'master')):
                filetype = 'ghost.redo'
            else:
                filetype = 'ghost.inverse'
        elif filetype == 'ghost with balloon':
            filetype = 'ghost.inverse'
        elif filetype in ['shell', 'shell with balloon']:
            if 'accept' in inst:
                filetype = 'supplement'
            else:
                filetype = 'shell.inverse'
        elif filetype == 'balloon':
            pass
        elif filetype == 'skin':
            filetype = 'nekoninni'
        elif filetype == 'katochan':
            pass
        else:
            fatal('unsupported file type(%s)' % filetype)
        if filetype in ['shell.inverse', 'ghost.inverse']:
            fatal('unsupported file type(%s)' % filetype)
        return filetype

    def install(self, filename, target, homedir):
        self.check_archive(filename)
        tmpdir = self.extract_files(filename)
        filetype = self.get_file_type(tmpdir)
        try:
            if filetype == 'ghost.redo':
                target_dir = self.install_redo_ghost(filename, tmpdir, homedir)
            elif filetype == 'supplement':
                target_dir = self.install_redo_supplement(filename, tmpdir, homedir, target)
            elif filetype == 'balloon':
                target_dir = self.install_balloon(filename, tmpdir, homedir)
            elif filetype == 'kinoko':
                self.install_kinoko(filename, tmpdir, homedir)
                target_dir = None # XXX
            elif filetype == 'nekoninni':
                self.install_nekoninni(filename, tmpdir, homedir)
                target_dir = None # XXX
            elif filetype == 'katochan':
                self.install_katochan(filename, tmpdir, homedir)
                target_dir = None # XXX
        finally:
            shutil.rmtree(tmpdir)
        return target_dir


    def download(self, url, basedir):
#        print 'downloading', url
        try:
            ifile = urllib.urlopen(url)
        except IOError:
            return None
        headers = ifile.info()
#        if 'content-length' in headers:
#            print '(size = %s bytes)' % headers.get('content-length')
        arcdir = ninix.home.get_archive_dir()
        if self.mode & DOWNLOAD: ## FIXME
            if not os.path.exists(arcdir):
                os.makedirs(arcdir)
        if not os.path.exists(arcdir):
            os.makedirs(arcdir)
        basedir = arcdir
        filename = os.path.join(basedir, os.path.basename(url))
        try:
            with open(filename, 'wb') as ofile:
                while 1:
                    data = ifile.read(4096)
                    if not data:
                        break
                    ofile.write(data)
        except IOError:
            return None
        ifile.close()
        # check the format of the downloaded file
        self.check_archive(filename) ## FIXME
        try:
            zf = zipfile.ZipFile(filename)
        except:
            return None
        test_zip = zf.testzip()
        zf.close()
        return None if test_zip is not None else filename

    def rename_files(self, basedir):
        if os.name == 'nt': # XXX
            return
        for filename in os.listdir(basedir):
            filename2 = filename.lower()
            path = os.path.join(basedir, filename2)
            if filename != filename2:
                os.rename(os.path.join(basedir, filename), path)
            if os.path.isdir(path):
                self.rename_files(path)

    def list_all_directories(self, top, basedir):
        dirlist = []
        for path in os.listdir(os.path.join(top, basedir)):
            if os.path.isdir(os.path.join(top, basedir, path)):
                dirlist.extend(self.list_all_directories(
                        top, os.path.join(basedir, path)))
                dirlist.append(os.path.join(basedir, path))
        return dirlist

    def remove_files_and_dirs(self, target_dir, mask):
        path = os.path.abspath(target_dir)
        if not os.path.isdir(path):
            return
        os.path.walk(path, self.remove_files, mask)
        dirlist = self.list_all_directories(path, '')
        dirlist.sort()
        dirlist.reverse()
        for name in dirlist:
            current_path = os.path.join(path, name)
            if os.path.isdir(current_path):
                head, tail = os.path.split(current_path)
                if tail not in mask and not os.listdir(current_path):
                    shutil.rmtree(current_path)

    def remove_files(self, mask, top_dir, filelist):
        for name in filelist:
            path = os.path.join(top_dir, name)
            if os.path.isdir(path) or name in mask:
                pass
            else:
                os.remove(path)

    def lower_files(self, top_dir):
        if os.name == 'nt': # XXX
            return    
        n = 0
        for filename in os.listdir(top_dir):
            filename2 = filename.lower()
            path = os.path.join(top_dir, filename2)
            if filename != filename2:
                os.rename(os.path.join(top_dir, filename), path)
                print 'renamed', os.path.join(top_dir, filename)
                n += 1
            if os.path.isdir(path):
                n += lower_files(path)
        return n

    def confirm(self, message):
        if self.dialog is not None:
            self.dialog.set_markup(message)
            response = self.dialog.run()
            self.dialog.hide()
            return response == gtk.RESPONSE_YES
        else:
            if not self.mode & INTERACTIVE:
                return True
            print ''.join(('Ninix Install:', message, '(yes/no)'))
            try:
                answer = raw_input()
            except EOFError:
                answer = None
            except KeyboardInterrupt:
                raise SystemExit
            if not answer:
                return False
            return answer.lower().startswith('y')

    def confirm_overwrite(self, path): ## FIXME: Use GUI
#        if not self.mode & INTERACTIVE:
#            return 1
#    print ''.join((PROGRAM_NAME, ': overwrite "%s"? (yes/no)' % path))
        return self.confirm('Overwrite "%s"?' % path)
#        self.dialog.set_markup('Overwrite "%s"?' % path)
#        response = self.dialog.run()
#        self.dialog.hide()
#        #print 'OVERWRITE:', response == gtk.RESPONSE_YES
#        return 1 if response == gtk.RESPONSE_YES else 0
#        try:
#            answer = raw_input()
#        except EOFError:
#            answer = None
#        except KeyboardInterrupt:
#            raise SystemExit
#        if not answer:
#            return 0
#        return answer.lower().startswith('y')

    def confirm_removal(self, path): ## FIXME: Use GUI
        return self.confirm('Remove "%s"?' % path)
#        if not self.mode & INTERACTIVE:
#            return 1
#        print 'INSTALL: remove "%s"? (yes/no)' % path
#        self.dialog.set_markup('Remove "%s"?' % path)
#        response = self.dialog.run()
#        self.dialog.hide()
#        #print 'REMOVE:', response == gtk.RESPONSE_YES
#        return 1 if response == gtk.RESPONSE_YES else 0
#        try:
#            answer = raw_input()
#        except EOFError:
#            answer = None
#        except KeyboardInterrupt:
#            raise SystemExit
#        if not answer:
#            return 0
#        return answer.lower().startswith('y')

    def install_redo_ghost(self, archive, tmpdir, homedir):
        # find install.txt
        inst = ninix.home.read_install_txt(tmpdir)
        if inst is None:
            fatal('install.txt not found')
        target_dir = inst.get('directory')
        if target_dir is None:
            fatal('"directory" not found in install.txt')
        target_dir = target_dir.encode('utf-8')
        prefix = os.path.join(homedir, 'ghost', target_dir)
        ghost_src = os.path.join(tmpdir, 'ghost', 'master')
        shell_src = os.path.join(tmpdir, 'shell')
        ghost_dst = os.path.join(prefix, 'ghost', 'master')
        shell_dst = os.path.join(prefix, 'shell')
        filelist = []
        ##filelist.append((os.path.join(tmpdir, 'install.txt'),
        ##                 os.path.join(prefix, 'install.txt'))) # XXX
        readme_txt = os.path.join(tmpdir, 'readme.txt')
        if os.path.exists(readme_txt):
            filelist.append((readme_txt,
                             os.path.join(prefix, 'readme.txt')))
        thumbnail_png = os.path.join(tmpdir, 'thumbnail.png')
        thumbnail_pnr = os.path.join(tmpdir, 'thumbnail.pnr')
        if os.path.exists(thumbnail_png):
            filelist.append((thumbnail_png,
                             os.path.join(prefix, 'thumbnail.png')))
        elif os.path.exists(thumbnail_pnr):
            filelist.append((thumbnail_pnr,
                             os.path.join(prefix, 'thumbnail.pnr')))
        for path in self.list_all_files(ghost_src, ''):
            filelist.append((os.path.join(ghost_src, path),
                             os.path.join(ghost_dst, path)))
        # find shell
        for path in self.list_all_files(shell_src, ''):
            filelist.append((os.path.join(shell_src, path),
                             os.path.join(shell_dst, path)))
        # find balloon
        balloon_dir = inst and inst.get('balloon.directory')
        if balloon_dir:
            balloon_list = []
            balloon_dir = ninix.home.get_normalized_path(balloon_dir)
            balloon_dst = os.path.join(homedir, 'balloon', balloon_dir)
            balloon_src = inst and inst.get('balloon.source.directory')
            if balloon_src:
                balloon_src = ninix.home.get_normalized_path(balloon_src)
            else:
                balloon_src = balloon_dir
            if os.path.exists(balloon_dst) and \
                    not self. confirm_removal(balloon_dst):
                pass # don't install balloon
            else:
                if os.path.exists(balloon_dst):
                    # uninstall older versions of the balloon
                    self.remove_files_and_dirs(balloon_dst, [])
                for path in self.list_all_files(tmpdir, balloon_src):
                    balloon_list.append((os.path.join(tmpdir, path),
                                     os.path.join(homedir, 'balloon', path)))
                self.install_files(balloon_list)
        if os.path.exists(prefix):
            inst_dst = ninix.home.read_install_txt(prefix)
            if not inst_dst or inst_dst.get('name') != inst.get('name'):
                self.mode |= INTERACTIVE ## FIXME
            if inst.get_with_type('refresh', int, 0):
                # uninstall older versions of the ghost
                if self.confirm_removal(prefix):
                    mask = [ninix.home.get_normalized_path(path) for path in \
                                inst.get('refreshundeletemask', '').split(':')]
                    mask.append('HISTORY')
                    self.remove_files_and_dirs(prefix, mask)
                else:
                    return
            else:
                if not self.confirm_overwrite(prefix):
                    return
        # install files
        print 'installing', archive, '(ghost)'
        self.install_files(filelist)
        # create SETTINGS
        path = os.path.join(prefix, 'SETTINGS')
        if not os.path.exists(path):
            try:
                with open(path, 'w') as f:
                    if balloon_dir:
                        f.write('balloon_directory, %s\n' % balloon_dir)
            except IOError, (code, message):
                logging.error('cannot write %s' % path)
        return target_dir

    def install_redo_supplement(self, archive, tmpdir, homedir, target):
        inst = ninix.home.read_install_txt(tmpdir)
        if inst and 'accept' in inst:
            print 'searching supplement target ...',
            candidates = []
            try:
                dirlist = os.listdir(os.path.join(homedir, 'ghost'))
            except OSError:
                dirlist = []
            for dirname in dirlist:
                path = os.path.join(homedir, 'ghost', dirname)
                if os.path.exists(os.path.join(path, 'shell', 'surface.txt')):
                    continue # ghost.inverse(obsolete)
                desc = ninix.home.read_descript_txt(
                    os.path.join(path, 'ghost', 'master'))
                if desc and desc.get('sakura.name') == inst.get('accept'):
                    candidates.append(dirname)
            if not candidates:
                print 'not found'
            elif target and target not in candidates:
                print "'%s' not found" % target
                return
            elif len(candidates) == 1 or target in candidates:
                if target:
                    path = os.path.join(homedir, 'ghost', target)
                else:
                    path = os.path.join(homedir, 'ghost', candidates[0])
                if 'directory' in inst:
                    if inst.get('type') == 'shell':
                        path = os.path.join(path, 'shell', inst['directory'])
                    else:
                        if 'type' not in inst:
                            print 'supplement type not specified'
                        else:
                            print 'unsupported supplement type:', inst['type']
                        return
                print 'found'
                if not os.path.exists(path):
                    os.makedirs(path)
                os.remove(os.path.join(tmpdir, 'install.txt'))
                distutils.dir_util.copy_tree(tmpdir, path)
                return target
            else: ## FIXME
                print 'multiple candidates found'
                for candidate in candidates:
                    print '   ', candidate
                fatal('try -S option with a target ghost name') ## FIXME

    def install_balloon(self, archive, srcdir, homedir):
        filelist = []
        # find install.txt
        inst = ninix.home.read_install_txt(srcdir)
        if inst is None:
            fatal('install.txt not found')
        target_dir = inst.get('directory')
        if target_dir is None:
            fatal('"directory" not found in install.txt')
        target_dir = target_dir.encode('utf-8')
        dstdir = os.path.join(homedir, 'balloon', target_dir)
        for path in self.list_all_files(srcdir, ''):
            filelist.append((os.path.join(srcdir, path),
                             os.path.join(dstdir, path)))
        ##filelist.append((os.path.join(srcdir, 'install.txt'),
        ##                 os.path.join(dstdir, 'install.txt')))
        if os.path.exists(dstdir):
            inst_dst = ninix.home.read_install_txt(dstdir)
            if not inst_dst or inst_dst.get('name') != inst.get('name'):
                self.mode |= INTERACTIVE
            if inst.get_with_type('refresh', int, 0):
                # uninstall older versions of the balloon
                if self.confirm_removal(dstdir):
                    mask = [ninix.home.get_normalized_path(path) for path in \
                                inst.get('refreshundeletemask', '').split(':')]
                    self.remove_files_and_dirs(dstdir, mask)
                else:
                    return
            else:
                if not self.confirm_overwrite(dstdir):
                    return
        # install files
        print 'installing', archive, '(balloon)'
        self.install_files(filelist)
        return target_dir

    def uninstall_kinoko(self, homedir, name):
        try:
            dirlist = os.listdir(os.path.join(homedir, 'kinoko'))
        except OSError:
            return
        for subdir in dirlist:
            path = os.path.join(homedir, 'kinoko', subdir)
            kinoko = ninix.home.read_kinoko_ini(path)
            if kinoko is None:
                continue
            kinoko_name = kinoko['title']
            if kinoko_name == name:
                kinoko_dir = os.path.join(homedir, 'kinoko', subdir)
                if self.confirm_removal(kinoko_dir):
                    shutil.rmtree(kinoko_dir)

    def install_kinoko(self, archive, srcdir, homedir):
        filelist = []
        # find kinoko.ini
        kinoko = ninix.home.read_kinoko_ini(srcdir)
        if kinoko is None:
            fatal('failed to read kinoko.ini')
        kinoko_name = kinoko['title']
        if kinoko['extractpath'] is not None:
            dstdir = os.path.join(
                homedir, 'kinoko',
                kinoko['extractpath'].encode('utf-8', 'ignore'))
        else:
            dstdir = os.path.join(
                homedir, 'kinoko', os.path.basename(archive)[:-4])
        # find files
        for filename in os.listdir(srcdir):
            path = os.path.join(srcdir, filename)
            if os.path.isfile(path):
                filelist.append((path, os.path.join(dstdir, filename)))
        # uninstall older versions of the kinoko
        self.uninstall_kinoko(homedir, kinoko_name)
        # install files
        print 'installing', archive, '(kinoko)'
        self.install_files(filelist)

    def uninstall_nekoninni(self, homedir, dir):
        nekoninni_dir = os.path.join(homedir, 'nekodorif', 'skin', dir)
        if not os.path.exists(nekoninni_dir):
            return
        if self.confirm_removal(nekoninni_dir):
            shutil.rmtree(nekoninni_dir)

    def install_nekoninni(self, archive, srcdir, homedir):
        filelist = []
        # find install.txt
        inst = ninix.home.read_install_txt(srcdir)
        if inst is None:
            fatal('install.txt not found')
        target_dir = inst.get('directory')
        if target_dir is None:
            fatal('"directory" not found in install.txt')
        target_dir = target_dir.encode('utf-8')
        dstdir = os.path.join(homedir, 'nekodorif', 'skin', target_dir)
        # find files
        for filename in os.listdir(srcdir):
            path = os.path.join(srcdir, filename)
            if os.path.isfile(path):
                filelist.append((path, os.path.join(dstdir, filename)))
        # uninstall older versions of the skin
        self.uninstall_nekoninni(homedir, target_dir)
        # install files
        print 'installing', archive, '(nekodorif skin)'
        self.install_files(filelist)

    def uninstall_katochan(self, homedir, target_dir):
        katochan_dir = os.path.join(
            homedir, 'nekodorif', 'katochan', target_dir)
        if not os.path.exists(katochan_dir):
            return
        if self.confirm_removal(katochan_dir):
            shutil.rmtree(katochan_dir)

    def install_katochan(self, archive, srcdir, homedir):
        filelist = []
        # find install.txt
        inst = ninix.home.read_install_txt(srcdir)
        if inst is None:
            fatal('install.txt not found')
        target_dir = inst.get('directory')
        if target_dir is None:
            fatal('"directory" not found in install.txt')
        target_dir = target_dir.encode('utf-8')
        dstdir = os.path.join(homedir, 'nekodorif', 'katochan', target_dir)
        # find files
        for filename in os.listdir(srcdir):
            path = os.path.join(srcdir, filename)
            if os.path.isfile(path):
                filelist.append((path, os.path.join(dstdir, filename)))
        # uninstall older versions of the skin
        self.uninstall_katochan(homedir, target_dir)
        # install files
        print 'installing', archive, '(nekodorif katochan)'
        self.install_files(filelist)

    def list_all_files(self, top, target_dir):
        filelist = []
        for path in os.listdir(os.path.join(top, target_dir)):
            if os.path.isdir(os.path.join(top, target_dir, path)):
                filelist.extend(self.list_all_files(
                        top, os.path.join(target_dir, path)))
            else:
                filelist.append(os.path.join(target_dir, path))
        return filelist

    def install_files(self, filelist):
        for from_path, to_path in filelist:
            dirname, filename = os.path.split(to_path)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            shutil.copy(from_path, to_path)
